// JSON validation
// written by LyricLy in 2021 for the 7th round of the Esolangs code guessing event

use std::borrow::Cow;
use std::collections::HashMap;
use std::str::Chars;
use std::vec;
use std::mem::ManuallyDrop;

#[derive(Copy, Clone, PartialEq, Eq, Debug)]
enum DelimType {
    Bracket,
    Brace,
}

#[derive(Clone, PartialEq, Debug)]
enum TokenType<'a> {
    Comma,
    Colon,
    Word(&'a str),
    String(Cow<'a, str>),
    Number(f64),
    OpenDelim(DelimType),
    CloseDelim(DelimType),
}

#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Debug)]
struct Span {
    lo: usize,
    hi: usize,
}

#[derive(Clone, PartialEq, Debug)]
struct Token<'a> {
    ty: TokenType<'a>,
    span: Span,
}

#[derive(Clone, PartialEq, Debug)]
enum Value<'a> {
    Bool(bool),
    Null,
    String(Cow<'a, str>),
    Number(f64),
    Array(Vec<Value<'a>>),
    Object(HashMap<Cow<'a, str>, Value<'a>>),
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
enum ExpectedTokenType {
    Value,
    Colon,
    StringOrClose(DelimType),
    ValueOrClose(DelimType),
    CommaOrClose(DelimType),
}

#[derive(Clone, PartialEq, Debug)]
enum DecodeErrorType<'a> {
    UnclosedDelim(DelimType),
    UnexpectedCloseDelim(DelimType),
    MismatchedCloseDelim(DelimType, DelimType),
    UnterminatedString,
    UnknownStartOfToken(char),
    UnexpectedToken(Token<'a>, ExpectedTokenType),
    UnexpectedEOF(ExpectedTokenType),
    UnknownEscape(char),
    InvalidUnicodeEscape(&'a str),
    ControlCharInString(char),
    InvalidObjectKey(Value<'a>),
    LeadingZero,
    LeadingPlus,
    NoIntegerPart,
    NoFractionalPart,
    NoExponent,
    ExtraData(Vec<Token<'a>>),
}

#[derive(Clone, PartialEq, Debug)]
struct DecodeError<'a> {
    ty: DecodeErrorType<'a>,
    span: Span,
}

// only support one error at a time because whatever
type Result<'a, T> = std::result::Result<T, DecodeError<'a>>;

#[derive(Debug, Clone)]
struct StringView<'a> {
    start: &'a str,
    iter: Chars<'a>,
}

impl<'a> Iterator for StringView<'a> {
    type Item = char;

    #[inline(always)]
    fn next(&mut self) -> Option<char> {
        self.iter.next()
    }

    #[inline(always)]
    fn count(self) -> usize {
        self.iter.count()
    }

    #[inline(always)]
    fn size_hint(&self) -> (usize, Option<usize>) {
        self.iter.size_hint()
    }

    #[inline(always)]
    fn last(self) -> Option<char> {
        self.iter.last()
    }
}

impl<'a> DoubleEndedIterator for StringView<'a> {
    #[inline(always)]
    fn next_back(&mut self) -> Option<char> {
        self.iter.next_back()
    }
}

impl<'a> StringView<'a> {
    fn pos(&self) -> usize {
        unsafe { (self.as_str().as_bytes().as_ptr()).offset_from(self.start.as_bytes().as_ptr()) as usize }
    }

    fn peek(&self) -> Option<char> {
        self.as_str().chars().next()
    }

    fn as_str(&self) -> &'a str {
        self.iter.as_str()
    }
}

struct Lexer<'a> {
    delims: Vec<(DelimType, Span)>,
    stream: StringView<'a>,
    last_ok_pos: usize,
}

impl<'a> Lexer<'a> {
    fn pos(&self) -> usize {
        self.stream.pos()
    }

    fn previous_pos_ok(&mut self) {
        self.last_ok_pos = self.pos();
    }

    fn span(&self) -> Span {
        Span { lo: self.last_ok_pos, hi: self.pos() }
    }

    fn error<T>(&self, ty: DecodeErrorType<'a>) -> Result<'a, T> {
        Err(DecodeError { ty, span: self.span() })
    }

    fn do_open_delim(&mut self, ty: DelimType) -> TokenType<'a> {
        self.delims.push((ty, self.span()));
        TokenType::OpenDelim(ty)
    }

    fn do_close_delim(&mut self, ty: DelimType) -> Result<'a, TokenType<'a>> {
        match self.delims.last() {
            None => self.error(DecodeErrorType::UnexpectedCloseDelim(ty)),
            Some(&(delim, _)) if delim == ty => {
                self.delims.pop();
                Ok(TokenType::CloseDelim(ty))
            },
            Some(&(delim, _)) => self.error(DecodeErrorType::MismatchedCloseDelim(delim, ty)),
        }
    }

    fn read_digits(&mut self) -> bool {
        let mut read_ok = false;
        loop {
            let c = self.stream.peek();
            if c.is_none() || !c.unwrap().is_ascii_digit() {
                break;
            }
            self.stream.next();
            read_ok = true;
        }
        read_ok
    }

    fn lex_json(s: &'a str) -> Result<'a, Vec<Token<'a>>> {
        let mut lexer = Lexer { stream: StringView { iter: s.chars(), start: s }, delims: Vec::new(), last_ok_pos: 0 };
        let mut tokens = Vec::new();

        while {
            while let Some(c) = lexer.stream.peek() {
                if !c.is_whitespace() {
                    break;
                }
                lexer.stream.next();
            }
            lexer.stream.peek().is_some()
        } {
            lexer.previous_pos_ok();
            let ps = lexer.stream.as_str();
            let s = &ps[1..];

            let ty = match lexer.stream.next().unwrap() {
                ',' => TokenType::Comma,
                ':' => TokenType::Colon,
                '{' => lexer.do_open_delim(DelimType::Brace),
                '}' => lexer.do_close_delim(DelimType::Brace)?,
                '[' => lexer.do_open_delim(DelimType::Bracket),
                ']' => lexer.do_close_delim(DelimType::Bracket)?,
                c if c.is_alphabetic() => {
                    let mut len = 1;
                    loop {
                        let c = lexer.stream.peek();
                        if c.is_none() || (!c.unwrap().is_alphanumeric() && c != Some('_')) {
                            break;
                        }
                        len += 1;
                        lexer.stream.next();
                    }
                    TokenType::Word(&ps[..len])
                },
                '"' => {  // "
                    let mut r = Cow::Borrowed(&s[..0]);
                    let mut len = 0;
                    loop {
                        len += 1;
                        let c = lexer.stream.next();
                        if c.is_none() {
                            return lexer.error(DecodeErrorType::UnterminatedString);
                        }
                        match c.unwrap() {
                            '"' => break,  // "
                            '\\' => {
                                let escape = lexer.stream.next();
                                len += 1;
                                if escape.is_none() {
                                    return lexer.error(DecodeErrorType::UnterminatedString);
                                }
                                let char = match escape.unwrap() {
                                    c@('"' | '\\' | '/') => c,  // "
                                    'b' => '\x08',
                                    'f' => '\x66',
                                    'n' => '\n',
                                    'r' => '\r',
                                    't' => '\t',
                                    'u' => {
                                        let slice = &s[len..(len+4).min(s.len()-1)];
                                        let slice = slice.split_once(&['\"', '\\'][..]).map(|(s, _)| s).unwrap_or(slice);  // "
                                        let n = u16::from_str_radix(slice, 16);
                                        if slice.len() < 4 || n.is_err() {
                                            return lexer.error(DecodeErrorType::InvalidUnicodeEscape(slice)).map_err(|mut err| {
                                                err.span.lo += len;
                                                err.span.hi += slice.len();
                                                err
                                            });
                                        }
                                        len += 4;
                                        for _ in 0..4 {
                                            lexer.stream.next();
                                        }
                                        char::from_u32(n.unwrap() as u32).unwrap()
                                    },
                                    c => return lexer.error(DecodeErrorType::UnknownEscape(c)).map_err(|mut err| {
                                        err.span.lo += len-1;
                                        err
                                    }),
                                };
                                r.to_mut().push(char);
                            },
                            c if !c.is_control() => match r {
                                Cow::Borrowed(_) => r = Cow::Borrowed(&s[..len]),
                                Cow::Owned(ref mut s) => s.push(c),
                            },
                            c => return lexer.error(DecodeErrorType::ControlCharInString(c)).map_err(|mut err| {
                                err.span.lo += len-1;
                                err
                            }),
                        }
                    }
                    TokenType::String(r)
                },
                c if c.is_ascii_digit() || c == '+' || c == '-' || c == '.' => {
                    // we just verify that this is a valid JSON number and use `parse` on it for simplicity
                    let mut error = match c {
                        '+' => Some(DecodeErrorType::LeadingPlus),
                        '.' => Some(DecodeErrorType::NoIntegerPart),
                        _ => None,
                    };

                    if c != '.' {
                        let saw_ok = lexer.read_digits();
                        if c == '0' && saw_ok {
                            error = error.or(Some(DecodeErrorType::LeadingZero));
                        } else if c == '-' && !saw_ok {
                            error = error.or(Some(DecodeErrorType::NoIntegerPart));
                        }
                    }

                    if c == '.' || (lexer.stream.peek() == Some('.')).then(|| ()).map(|_| lexer.stream.next()).is_some() {
                        if !lexer.read_digits() {
                            error = error.or(Some(DecodeErrorType::NoFractionalPart));
                        }
                    }

                    if let Some('e' | 'E') = lexer.stream.peek() {
                        lexer.stream.next();
                        if let Some('+' | '-') = lexer.stream.peek() {
                            lexer.stream.next();
                        }
                        if !lexer.read_digits() {
                            error = error.or(Some(DecodeErrorType::NoExponent));
                        }
                    }

                    if let Some(err) = error {
                        return lexer.error(err);
                    }
                    TokenType::Number(ps[..lexer.pos()-lexer.last_ok_pos].parse().unwrap())
                },
                c => return lexer.error(DecodeErrorType::UnknownStartOfToken(c)),
            };

            tokens.push(Token { ty, span: lexer.span() } );
        }

        if !lexer.delims.is_empty() {
            return Err(DecodeError { ty: DecodeErrorType::UnclosedDelim(lexer.delims[0].0), span: lexer.delims[0].1 });
        }

        Ok(tokens)
    }
}


// there is absolutely no reason to do this this way but I feel like it'd be a waste if I didn't write something cursed at least once in this file
#[derive(Debug)]
struct MyIter<T> {
    buf: std::ptr::NonNull<T>,
    cap: usize,
    ptr: *const T,
    end: *const T,
}

fn deprivatize<'a>(i: &vec::IntoIter<Token>) -> &'a MyIter<Token<'a>> {
    unsafe { std::mem::transmute(i) }
}

fn peek_ptr<'a>(i: &vec::IntoIter<Token>) -> *const Token<'a> {
    deprivatize(i).ptr
}

fn parse_value<'a>(i: &mut vec::IntoIter<Token<'a>>, exp: ExpectedTokenType) -> Result<'a, Value<'a>> {
    let tok = i.next();
    if tok.is_none() {
        let di = deprivatize(i);
        let span = if di.buf.as_ptr() as *const _ != di.ptr {
            let last = unsafe { &*di.ptr.offset(-1) };
            Span { lo: last.span.hi+1, hi: last.span.hi+2 }
        } else {
            Span { lo: 0, hi: 1 }
        };
        return Err(DecodeError { ty: DecodeErrorType::UnexpectedEOF(exp), span });
    }
    let tok = tok.unwrap();
    let r = match tok.ty {
        TokenType::Number(n) => Value::Number(n),
        TokenType::String(s) => Value::String(s),
        TokenType::Word(w) => match w {
            "true" => Value::Bool(true),
            "false" => Value::Bool(false),
            "null" => Value::Null,
            _ => return Err(DecodeError { span: tok.span, ty: DecodeErrorType::UnexpectedToken(tok, exp) }),
        },
        TokenType::OpenDelim(DelimType::Bracket) => {
            let mut arr = Vec::new();
            match unsafe { &(*peek_ptr(i)).ty } {
                TokenType::CloseDelim(DelimType::Bracket) => {
                    i.next();
                    return Ok(Value::Array(arr));
                },
                _ => arr.push(parse_value(i, ExpectedTokenType::ValueOrClose(DelimType::Bracket))?),
            }
            loop {
                // can't be None because we already know brackets are balanced
                let tok = i.next().unwrap();
                match tok.ty {
                    TokenType::CloseDelim(DelimType::Bracket) => break,
                    TokenType::Comma => arr.push(parse_value(i, ExpectedTokenType::Value)?),
                    _ => return Err(DecodeError { span: tok.span, ty: DecodeErrorType::UnexpectedToken(tok, ExpectedTokenType::CommaOrClose(DelimType::Bracket)) }),
                }
            }
            Value::Array(arr)
        },
        TokenType::OpenDelim(DelimType::Brace) => {
            let mut obj = HashMap::new();
            match unsafe { &(*peek_ptr(i)).ty } {
                TokenType::CloseDelim(DelimType::Brace) => {
                    i.next();
                    return Ok(Value::Object(obj));
                },
                _ => {
                    let key_start_tok = unsafe { &*peek_ptr(i) };
                    let key = parse_value(i, ExpectedTokenType::StringOrClose(DelimType::Brace))?;
                    let key_end_tok = unsafe { &*peek_ptr(i).offset(-1) };
                    let tok = i.next().unwrap();
                    if let TokenType::Colon = tok.ty {
                    } else {
                        return Err(DecodeError { span: tok.span, ty: DecodeErrorType::UnexpectedToken(tok, ExpectedTokenType::Colon) });
                    }
                    let val = parse_value(i, ExpectedTokenType::Value)?;
                    if let Value::String(s) = key {
                        obj.insert(s, val);
                    } else {
                        return Err(DecodeError { span: Span { lo: key_start_tok.span.lo, hi: key_end_tok.span.hi }, ty: DecodeErrorType::InvalidObjectKey(key) });
                    }
                },
            }
            loop {
                // as above
                let tok = i.next().unwrap();
                match tok.ty {
                    TokenType::CloseDelim(DelimType::Brace) => break,
                    TokenType::Comma => {
                        let key_start_tok = unsafe { &*peek_ptr(i) };
                        let key = parse_value(i, ExpectedTokenType::Value)?;
                        let key_end_tok = unsafe { &*peek_ptr(i).offset(-1) };
                        let tok = i.next().unwrap();
                        if let TokenType::Colon = tok.ty {
                        } else {
                            return Err(DecodeError { span: tok.span, ty: DecodeErrorType::UnexpectedToken(tok, ExpectedTokenType::Colon) });
                        }
                        let val = parse_value(i, ExpectedTokenType::Value)?;
                        if let Value::String(s) = key {
                            obj.insert(s, val);
                        } else {
                            return Err(DecodeError { span: Span { lo: key_start_tok.span.lo, hi: key_end_tok.span.hi }, ty: DecodeErrorType::InvalidObjectKey(key) });
                        }
                    },
                    _ => return Err(DecodeError { span: tok.span, ty: DecodeErrorType::UnexpectedToken(tok, ExpectedTokenType::CommaOrClose(DelimType::Brace)) }),
                }
            }
            Value::Object(obj)
        }
        _ => return Err(DecodeError { span: tok.span, ty: DecodeErrorType::UnexpectedToken(tok, exp) }),
    };

    Ok(r)
}

fn parse_json<'a>(s: &'a str) -> Result<'a, Value<'a>> {
    let mut i = Lexer::lex_json(s)?.into_iter();
    let r = parse_value(&mut i, ExpectedTokenType::Value)?;
    // lmao
    let rest = i.as_slice();
    let r = if !rest.is_empty() {
        unsafe {
            let span = Span { lo: rest[0].span.lo, hi: rest[rest.len()-1].span.hi };
            let v = deprivatize(&i);
            let len = v.end.offset_from(v.ptr) as usize;
            std::ptr::copy(v.ptr, v.buf.as_ptr(), len);
            Err(DecodeError { span, ty: DecodeErrorType::ExtraData(Vec::from_raw_parts(v.buf.as_ptr(), len, v.cap)) })
        }
    } else {
        Ok(r)
    };
    let mut i = ManuallyDrop::new(i);
    if let Ok(_) = r {
        unsafe { ManuallyDrop::drop(&mut i) };
    }
    r
}

// all done
pub fn entry(s: &str) -> bool {
    matches!(parse_json(s), Ok(_))
}
